function [ NA_LG, delta_d_i ] = m_gen_PSF_aperture_func_focal_position( params )
% This function calculates the amplitude of electric field and the
% aberration of the wavefront and returns the complex the electric field at
% the tube lens' aperture plane.

% PARENT: m_gen_PSF_numerical_calc.m
% CHILD: m_gen_PSF_aberration.m

int_method='custom';
%int_method='Simpson';
z_correction_method='variance_delta_d_i';
%z_correction_method='empirical';

aperture_func_bool=false;

psf_type=params.psf.type;
lambda=params.dye.lambda*1E-9;  % emission wavelength in m
k=2*pi/lambda;   % wave number in "1/m"

switch psf_type
    case 'scalar'
        %PSF_model='custom_obj';            % Fourier transform of electric field at the microscope's frontal aperture
        %PSF_model='custom_tube';           % Fourier transform of electric field at tube lens' aperture
        %PSF_model='Airy';                  % calculates Airy diffraction pattern
        %PSF_model='Gibson and Lanni';       % Airy diffraction including Gibson and Lanni aberration term
        %PSF_model='Pankajakshan';          % Gibson and Lanni PSF model including the apodizing function (sqrt(cos(theta)))
        PSF_model='modified Gibson and Lanni';       % Gibson and Lanni PSF model including the apodizing function and transmission coefficients
    case 'vectorial'
        %PSF_model='custom_obj';
        %PSF_model='custom_tube';
        PSF_model='Richards and Wolf';     % Richards and Wolf PSF model
end

% sources:
% Gibson and Lanni:     "Experimental test of an analytical model of aberration in an oil-immersion objective lens used in three-dimensional light microscopy"
%                       "Focusing of light through a stratified medium: a practical approach for computing microscope point spread functions. Part I: Conventional microscopy"
% Pankajakshan:         "Point-spread function retrieval for fluorescence microscopy"
% Richards and Wolf:    "Super-resolution orientation estimation and localization of fluorescent dipoles using 3-D steerable filters"
% T�r�k and Varga:      "Focusing of light through a stratified medium: a practical approach for computing microscope point spread functions. Part I: Conventional microscopy"
%                       "Is it differs from the "Richards and Wolf" model?"
% Haeberl�:             "Focusing of light through a stratified medium: a practical approach for computing microscope point spread functions. Part I: Conventional microscopy"
%                       "Fluorescence microscopy point spread function model accounting for aberrations due to refractive index variability within a specimen"
%                       "Is it differs from the "Richards and Wolf" model?"

switch PSF_model
    case 'custom_tube'
        Abbe_bool=true;     % boolean for whether the intesity sholund be distorted between the objective's and tube lens' aperture according to ray distribution following Abbe sine condition under desing conditions
    otherwise
        Abbe_bool=false;     % boolean for whether the intesity sholund be distorted between the objective's and tube lens' aperture according to ray distribution following Abbe sine condition under desing conditions
end

if strcmp(psf_type,'vectorial')
    anisotropy_bool=params.psf.anisotropy_bool;
    orientation=params.mol.orientation;
end

aperture_rad_tube=params.numcalc.aperture_rad_tube;     % modified radius of tube lens' aperture in m
aperture_rad_obj=params.numcalc.aperture_rad_obj;       % radius of objective's frontal aperture in m
focal_plane_dist=params.numcalc.focal_plane_dist;       % distance between the sample and the objective's aperture under desing conditions



    function [x, functionValue, exitflag] = PSF_focus_iteration_custom(functionHandle, x0)
        
        % parameters for the iteration
        initialStepSize=100*1e-9;
        derivatStep=1*1e-9;
        stepCriterion=1*1e-9;
        
        % calculate the derivate at the initial "x":
        f1=functionHandle(x0-derivatStep/2);
        f2=functionHandle(x0+derivatStep/2);
        functionDerivate=(f2-f1)/derivatStep;
        initialDerivate=functionDerivate;
        initialValue=(f1+f2)/2;
        
        x=x0;
        step=initialStepSize;
        while abs(step)>stepCriterion
            
            f1=functionHandle(x-derivatStep/2);
            f2=functionHandle(x+derivatStep/2);
            
            functionDerivate=(f2-f1)/derivatStep;
            
            step=-functionDerivate/initialDerivate*initialStepSize;
            
            x=x+step;
            
        end
        
        
        f1=functionHandle(x-derivatStep/2);
        f2=functionHandle(x+derivatStep/2);
        functionValue=(f1+f2)/2;
        
        if functionValue<initialValue
            exitflag=1;
        elseif functionValue==initialValue
            exitflag=1;
        else
            exitflag=-1;
        end
        
    end



% function for varying theta angle in the immersion medium in order that
% the ray hit the sampled point on the aperture
    function F=ray_angles(theta, h_ap)
        F=(h_ap-(d_i+delta_d_i)*tan(theta)-...
            d_g*n_i*sin(theta)/n_g./sqrt(1-(n_i*sin(theta)/n_g).^2)-...
            d_s*n_i*sin(theta)/n_s./sqrt(1-(n_i*sin(theta)/n_s).^2))*1E9;
    end

    function [theta, error_ray_angles_iteration, exitflag_ray_angles_iteration]=ray_angles_iteration(theta_edge_i_0, theta_i_max_0, theta_i_min_0, h_ap)
        theta=theta_edge_i_0;
        theta_max=theta_i_max_0;        % maximal angle value, the angle belonging to the edge must be under it
        theta_min=theta_i_min_0;        % minimal angle value, the angle belonging to the edge must be over it
        exitflag_ray_angles_iteration=0;
        delta_x_ap_0=-ray_angles(theta, h_ap);  % initial position change where the ray intersects the aperture
        theta_prev=theta;   % theta of the previous iteration
        for idxTheta=1:100
            theta_prev2=theta_prev; % theta of the 2 previous iteration
            theta_prev=theta;
            delta_x_ap=-ray_angles(theta, h_ap);    % position change where the ray intersects the aperture
            if delta_x_ap<0                         % if the ray intersects the aperture below the edge
                theta_min=theta;
                theta=(5*theta_max+theta_min)/6;    % weighted mean, for faster convergence
            elseif delta_x_ap>0                     % if the ray intersects the aperture above the edge
                theta_max=theta;
                theta=(5*theta_min+theta_max)/6;    % weighted mean, for faster convergence
            elseif delta_x_ap==0                    % if the ray intersects the aperture exactly at the edge
                exitflag_ray_angles_iteration=1;
                error_ray_angles_iteration=(delta_x_ap)/h_ap;
                break;
            end
            
            % if the "theta" change is in the order of the smallest number,
            % without it, it might not converge to it, only oscillate
            % around it
            if abs(theta-theta_prev)==eps
                if abs(delta_x_ap)<=h_ap*1e-4
                    exitflag_ray_angles_iteration=1;
                    break;
                end
            elseif theta==theta_prev2
                if theta>theta_prev
                    theta=theta_prev+eps;
                elseif theta<theta_prev
                    theta=theta_prev-eps;
                end
            end
            
        end
        
        % in this case, consider the iteration successful
        if abs(delta_x_ap)<=h_ap*1e-4 
            exitflag_ray_angles_iteration=1;
        end
        error_ray_angles_iteration=(delta_x_ap)/h_ap;
    end

% error function of the surface (defocus on unit circle) fitting
    function G=defocus_fitting(fit_var)
        G=sum(sum(abs((Lanni_Gibson_ab_orig-(fit_var(1)+fit_var(2)*ro_samp_tube_2)).*circ).^error_power.*error_weights.*int_mat))*tolerance_const;
    end

% function for calculating the radial dependence of PSF during PSF
% size minimalization of PSF distorted by Gibson and Lanni aberration
%     function H=PSF_func(delta_d_i_iter, defocus_iter)
%          H=abs(...
%              (besselj(0,2*pi*2.5*0.5*ro_det_var*ro_tube_var')*...
%              (ro_tube_var.*apodizing_func_PSF_variance...           % Hankel transform of the approximated aperture function
%              .*exp(1i*k*(d_s*sqrt(n_s^2-NA_LG^2*ro_tube_var.^2)+delta_d_i_iter*sqrt(n_i^2-NA_LG^2*ro_tube_var.^2)-defocus_iter*ro_tube_var.^2)))...
%              )'...
%          ).^2;    % Gibson and Lanni aberration
%     end

    function H=PSF_func(delta_d_i_iter, defocus_iter)
         H=abs(...
             (besselj(0,2*pi*2.5*0.5*ro_det_var*ro_tube_var')*...
             (ro_tube_var.*1 ...           % Hankel transform of the approximated aperture function
             .*exp(1i*k*(d_s*sqrt(n_s^2-NA_LG^2*ro_tube_var.^2)+delta_d_i_iter*sqrt(n_i^2-NA_LG^2*ro_tube_var.^2)-defocus_iter*ro_tube_var.^2)))...
             )'...
         ).^2;    % Gibson and Lanni aberration
    end


% FOR testing: no apodization or higher order aberration:
%     function H=PSF_func(delta_d_i_iter, defocus_iter)
%          H=abs(...
%              (besselj(0,2*pi*2.5*0.5*ro_det_var*ro_tube_var')*...
%              (ro_tube_var.*1 ...           % Hankel transform of the approximated aperture function
%              .*exp(1i*k*(d_s*n_s*(1-0.5*NA_LG^2*ro_tube_var.^2/n_s^2)+delta_d_i_iter*n_i*(1-0.5*NA_LG^2*ro_tube_var.^2/n_i^2)-defocus_iter*ro_tube_var.^2)))...
%              )'...
%          ).^2;    % Gibson and Lanni aberration
%     end

% function for calculating the variance of PSF during PSF
% size minimalization of PSF disrtorted by Gibson and Lanni aberration
    function I=PSF_variance(delta_d_i_iter)
        defocus_iter=0;                                 % do not add defocus arbitrarily
        I=PSF_func(delta_d_i_iter, defocus_iter)...                           % PSF
            /(PSF_func(delta_d_i_iter, defocus_iter)*2*pi*ro_det_var)...     % normalization
            *(2*pi*ro_det_var.^3)*1E3;                             % multipliing with radius^2 for variance and integrating in polar coordiantes
    end

% function for calculating the variance of PSF during PSF
% size minimalization of PSF disrtorted by Gibson and Lanni aberration
    function J=PSF_defocus_variance(defocus_iter)
        delta_d_i_iter=delta_d_i;
        J=PSF_func(delta_d_i_iter, defocus_iter)...                         % PSF
            /(PSF_func(delta_d_i_iter, defocus_iter)*2*pi*ro_det_var)...     % normalization
            *(2*pi*ro_det_var.^3)*1E3;                         % multipliing with radius^2 for variance and integrating in polar coordiantes
        
    end

    function apodizing_func_approx=apodizing_func_approx_cacl(theta_edge_i, h_ap, PSF_model)
        % Approximating the apodization and reflections to get an
        % approximate normalized electic field values on the
        % objective's aperture.
        % Lets consider the source is in a homogenous medium in the
        % aspect of the propagation direction, so no iteration has
        % to used to calculate the exact ray angles
        
        % lets calculate source's distance to the objective's
        % aperture from the already calculated marginal ray angle
        d_approx=h_ap/tan(theta_edge_i);
        
        % calculate the ray angles in every mediums to be able
        % to consider the transmission coefficients and the
        % apodization effect
        theta_i_approx=atan(h_ap*ro_tube_var/d_approx);
        theta_g_approx=asin(n_i/n_g*sin(theta_i_approx));
        theta_s_approx=asin(n_i/n_s*sin(theta_i_approx));
        
        % transmission coefficient of different surfaces for individual rays
        t_s_sg_approx=2*n_s*cos(theta_s_approx)./(n_s*cos(theta_s_approx)+n_g*cos(theta_g_approx));
        t_p_sg_approx=2*n_s*cos(theta_s_approx)./(n_s*cos(theta_g_approx)+n_g*cos(theta_s_approx));
        t_s_gi_approx=2*n_g*cos(theta_g_approx)./(n_g*cos(theta_g_approx)+n_i*cos(theta_i_approx));
        t_p_gi_approx=2*n_g*cos(theta_g_approx)./(n_g*cos(theta_i_approx)+n_i*cos(theta_g_approx));
        
        % in case of total internal reflection
        t_s_sg_approx(imag(t_s_sg_approx)~=0)=0;
        t_p_sg_approx(imag(t_p_sg_approx)~=0)=0;
        t_s_gi_approx(imag(t_s_gi_approx)~=0)=0;
        t_p_gi_approx(imag(t_p_gi_approx)~=0)=0;
        
        % total transmission coefficients for the individual rays
        t_s_approx=t_s_sg_approx.*t_s_gi_approx;
        t_p_approx=t_p_sg_approx.*t_p_gi_approx;
        
        t_approx=(t_s_approx+t_p_approx)/2;     % the transmission coefficient
        apodization_approx=sqrt(cos(theta_i_approx));   % the apodization effect
        E_aperture_change_approx=d_approx./sqrt(((h_ap*ro_tube_var).^2+d_approx^2));     % electic field deviation between the middle and marginal part of the aperture because of the different distance from the source
        
        % the overall apodization of the electric field
            switch PSF_model
                case {'custom_obj','custom_tube'}
                    apodizing_func_approx=t_approx.*apodization_approx.*E_aperture_change_approx;
                case {'Airy', 'Gibson and Lanni'}
                    apodizing_func_approx=ones(size(ro_tube_var));
                case {'Pankajakshan'}
                    apodizing_func_approx=apodization_approx;
                case {'modified Gibson and Lanni', 'Richards and Wolf'}
                    apodizing_func_approx=t_approx.*apodization_approx;
            end
    end


%ray_angle_iteration_method='matlab';
ray_angle_iteration_method='custom';

%apodization_approx_method='empirical';
apodization_approx_method='physical approx';

% Calculating constant functions which do not change from molecule to molecule
if ~aperture_func_bool      % if it has not been calculated for the actual dye or the parameters has been changed (generate button has benn pressen once again)
    
    % reading constants
    em_N_ofphotonpersec_mean=params.dye.em_N_ofphotonpersec;        % avarage number of emitted photons per second (avarage of emissions with all possible dipole orientations)
    NA=params.aq.NA;
    sampling_number_x=params.numcalc.sampling_number_x;
    sampling_number_y=params.numcalc.sampling_number_y;
    sampling_index_x=params.numcalc.sampling_index_x;
    sampling_index_y=params.numcalc.sampling_index_y;
    aperture_sampling_fill_bool=params.numcalc.aperture_sampling_fill_bool;
    
    n_i=params.aq.n_immersion;      % refractive index of immersion medium
    n_g=n_i;                        % refractive index of cover slip
    n_s=params.aq.n_sample;        % refractive index of sample
    
    d_i=focal_plane_dist-params.numcalc.z_focal_x;	% thickness of the immersion medium (and the cover slip) under design conditions in m
    d_g=0*1E-6;                          % thickness of the cover slip, it's taken into account in "d_i" in m
    % The aberration because the shift of joint focal point in case of 3D
    % astigmatic measurement is not taken into account because "d_g" is
    % zero and otherwise it would be insignificant in normal case.
    d_s=params.psf.sample_depth*1E-6;          % thickness of the sample layer in m
    
    % determining the maximal ray angles belonging to the total intrenal
    % refraction in the immersion medium
    switch max([n_s, n_g, n_i])
        case n_i    % if the immersion medium has the highest refractive index
            theta_i_max=asin(min([n_s, n_g, n_i])/n_i);
        case n_g    % if the cover slip has the highest refractive index
            if n_i>n_s
                theta_i_max=asin(min([n_s, n_g, n_i])/n_i);
            elseif n_i<=n_s
                theta_i_max=pi/2;
            end
        case n_s    % if the sample medium has the highest refractive index
            if n_i>n_g
                theta_i_max=asin(min([n_s, n_g, n_i])/n_i);
            elseif n_i<=n_g
                theta_i_max=pi/2;
            end
    end
    
    switch PSF_model
        case {'Airy'}
            delta_d_i=-d_s*n_i/n_s;
        otherwise
            switch z_correction_method
                case 'variance_delta_d_i'
                    % Minimizing the PSF variance with immersion thickness change.
                    sampling_number_var=512;
                    
                    ro_tube_var=(0.5:sampling_number_var-0.5)'/sampling_number_var;        % normalized radii values of the tube lens' aperture
                    ro_det_var=(0.5:sampling_number_var-0.5)'/sampling_number_var*2;         % normalized radii values of the imaged area
                    
                    delta_d_i_0=-d_s*n_i/n_s;               % initial thickness change of the immersion medium
                    delta_d_i=delta_d_i_0;                  % thickness change of the immersion medium
                    
                    theta_max_ratio=0.95;   % starting maximum allowed ray angle and angle of total internal reflection ratio
                    % initial guess for the ray angles belonging to the intersection border points
                    theta_edge_i_0=min(atan(abs(aperture_rad_obj)/(d_i+delta_d_i+d_s)), theta_i_max*theta_max_ratio);     % initial guess for the ray angles, required for the iteration method
                    
                    for var_iter_idx=1:1        % let this the maximal number of iterations for calculating the "z" pos where the PSF variance is minimal
                        var_iter_success_bool=true;    % do not break out of the "while" loop
                        % running the itarition with different "theta_bord_i_0_max" values
                        
                        switch ray_angle_iteration_method
                            case 'matlab'
                                options = optimset('Display','off', 'FinDiffType', 'central');        % do not display iteration output
                                N_fsolve_loop=10;       % maximal number of loops for the iteration of "theta_i_ap_edge", inner loops, should be enough
                                fsolve_loop=1;       % variable for the "while" loop
                                while fsolve_loop<=N_fsolve_loop
                                    iter_success_bool=true;    % boolean for storing whether all the iterations were successful
                                    theta_edge_i_0_max=theta_i_max*theta_max_ratio;            % let this be the maximal initial ray angle, should be smaller than the angle of total internal refraction
                                    % initial ray angle for the iteration
                                    if var_iter_idx==1
                                        theta_edge_i_0_iter=theta_edge_i_0;
                                    else
                                        theta_edge_i_0_iter=theta_edge_i;
                                    end
                                    theta_edge_i_0_iter(theta_edge_i_0_iter>theta_edge_i_0_max)=theta_edge_i_0_max;   % if any of the initial guess exceeds the maximal value...
                                    
                                    [theta_edge_i, ~, exitflag]=fsolve(@(theta) ray_angles(theta, abs(aperture_rad_obj)), theta_edge_i_0_iter, options);   % iteration for getting the ray angles
                                    % examining whether the iteration was successfull for a given ray
                                    if exitflag>0       % if the iteration was successful
                                    else                % if the iteration was not successful
                                        iter_success_bool=false;    % do not break out of the "while" loop
                                    end
                                    
                                    % if the iterations were successful for all angles...
                                    if iter_success_bool
                                        break;           % break out of the loop
                                    end
                                    % decrease the "theta_bord_i_0_max" and "theta_i_max" difference in
                                    % the next loop
                                    fsolve_loop=fsolve_loop+1;
                                    theta_max_ratio=1-(1-theta_max_ratio)/2;
                                end
                            case 'custom'
                                [theta_edge_i, ~, exitflag]=ray_angles_iteration(theta_edge_i_0, theta_i_max, 0, abs(aperture_rad_obj));
                                % examining whether the iteration was successfull for a given ray
                                if exitflag>0       % if the iteration was successful
                                    iter_success_bool=true;
                                else                % if the iteration was not successful
                                    iter_success_bool=false;
                                end
                        end
                        if ~iter_success_bool
                            theta_edge_i=theta_i_max;
                            %error('Could not calculate ray angle for aperture edge. Please decrease the numerical aperture.')
                        end
                        
                        
                        
                        % approximated apodizing function, gives the "shape" of the aperture function
                        % did not want to calculate it here, did not want to run the expensive
                        % iteration for all ray angles here too and experienced as the the shape of
                        % the aperture function does not affects much the outcome variable value of
                        % minimalization of the PSF
                        switch apodization_approx_method
                            case 'empirical'
                                apodizing_func_PSF_variance=cos(theta_edge_i*ro_tube_var).^1.5;           % this fits relatively well to the product of the apodizing function and the transmission coefficients (modified apodizing function)
                                %apodizing_func=cos(atan(abs(aperture_rad_obj*ro_tube_iter)/(delta_d_i))).^0.5;
                                %apodizing_func_PSF_variance=1;
                            case 'physical approx'
                                apodizing_func_PSF_variance=apodizing_func_approx_cacl(theta_edge_i, abs(aperture_rad_obj), PSF_model);
                        end
                        
                        
                        NA_LG=n_i*sin(theta_edge_i);    % numerical aperture in Gibson and Lanni aberration
                        
                        
                        delta_d_i_0=delta_d_i;      % starting immersion thickness change for the PSF variance calculation
                        
                        %[delta_d_i,~,exitflag_var_iter]=fminsearch(@(delta_d_i_iter) PSF_variance(delta_d_i_iter), delta_d_i_0);       % minimizing the PSF variance
                        [delta_d_i,~,exitflag_var_iter]=fminsearch(@(delta_d_i_iter) PSF_variance(delta_d_i_iter), -6.5*10^-6);
                        %[delta_d_i,~,exitflag_var_iter]=PSF_focus_iteration_custom(@(delta_d_i_iter) PSF_variance(delta_d_i_iter), delta_d_i_0);       % minimizing the PSF variance
                        
                        H=PSF_func(delta_d_i, 0);
                        figure()
                        plot(H)
                        
                        if exitflag_var_iter>0       % if the iteration was successful
                            if abs((delta_d_i-delta_d_i_0)/delta_d_i_0)<1E-4
                                var_iter_success_bool=true;
                                break;
                            end
                        else                % if the iteration was not successful
                            var_iter_success_bool=false;    % break out of the "for" loop
                            break;
                        end
                        
                    end
                    
                    if ~var_iter_success_bool       % if the iteration was successful
                        error('Could not minimalize the PSF variance. Please decrease the numerical aperture or the simple medium thickness.')                % if the iteration was not successful
                    end
                case 'variance_defocus'
                    % Minimizing the PSF variance with defocus
                    
                    delta_d_i=-d_s*n_i/n_s;
                case 'empirical'
                    % Minimizing the PSf width with defocus determined with
                    % empirical constants.
                    
                    delta_d_i=-d_s*n_i/n_s;
            end
    end
    

    
end

NA_LG=n_i*sin(theta_edge_i);    % numerical aperture in Gibson and Lanni aberration

end
